//! # Changelog v0.3
//!
//! * Utf8 type is nullable [#37](https://github.com/ritchie46/polars/issues/37)
//! * Support all ARROW numeric types [#40](https://github.com/ritchie46/polars/issues/40)
//! * Support all ARROW temporal types [#46](https://github.com/ritchie46/polars/issues/46)
//! * ARROW IPC Reader/ Writer [#50](https://github.com/ritchie46/polars/issues/50)
//! * Implement DoubleEndedIterator trait for ChunkedArray's [#34](https://github.com/ritchie46/polars/issues/34)
//!
